# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from .show_versions import show_versions  # noqa: F401


__name__ = "pyrit"
# Remove dev suffix when releasing and keep in sync with pyproject.toml
__version__ = "0.4.0"
