from setuptools import setup, find_packages

setup(
    name='pyrit-library',
    version='0.1.5',
    packages=find_packages(),
    package_data={
        '': [
            'prompt_chat_target/*',
            'datasets/*/*.yaml',
            'datasets/*/*/*.yaml',
            ],
    },
    install_requires=[
        'aioconsole',
        'annotated-types',
        'anyio',
        'appdirs',
        'argon2-cffi',
        'argon2-cffi-bindings',
        'arrow',
        'art',
        'asttokens',
        'async-lru',
        'attrs',
        'azure-ai-contentsafety',
        'azure-ai-ml',
        'azure-cognitiveservices-speech',
        'azure-common',
        'azure-core',
        'azure-identity',
        'azure-mgmt-core',
        'azure-storage-blob',
        'azure-storage-file-datalake',
        'azure-storage-file-share',
        'babel',
        'beautifulsoup4',
        'bleach',
        'cachetools',
        'certifi',
        'cffi',
        'charset-normalizer',
        'colorama',
        'coloredlogs',
        'comm',
        'confusables',
        'cryptography',
        'debugpy',
        'decorator',
        'defusedxml',
        'distro',
        'duckdb',
        'duckdb_engine',
        'executing',
        'fastjsonschema',
        'filelock',
        'flatbuffers',
        'fqdn',
        'fsspec',
        'google-api-core',
        'google-auth',
        'googleapis-common-protos',
        'greenlet',
        'h11',
        'httpcore',
        'httpx',
        'huggingface-hub',
        'humanfriendly',
        'idna',
        'ipykernel',
        'ipython',
        'ipywidgets',
        'isodate',
        'isoduration',
        'jedi',
        'Jinja2',
        'jiter',
        'joblib',
        'json5',
        'jsonpickle',
        'jsonpointer',
        'jsonschema',
        'jsonschema-specifications',
        'jupyter',
        'jupyter-console',
        'jupyter-events',
        'jupyter-lsp',
        'jupyter_client',
        'jupyter_core',
        'jupyter_server',
        'jupyter_server_terminals',
        'jupyterlab',
        'jupyterlab_pygments',
        'jupyterlab_server',
        'jupyterlab_widgets',
        'logzero',
        'MarkupSafe',
        'marshmallow',
        'matplotlib-inline',
        'mistune',
        'mpmath',
        'msal',
        'msal-extensions',
        'msrest',
        'mypy',
        'mypy-extensions',
        'nbclient',
        'nbconvert',
        'nbformat',
        'nest-asyncio',
        'notebook',
        'notebook_shim',
        'numpy',
        'oauthlib',
        'onnx',
        'onnxruntime',
        'openai',
        'opencensus',
        'opencensus-context',
        'opencensus-ext-azure',
        'overrides',
        'packaging',
        'pandocfilters',
        'parso',
        'pillow',
        'platformdirs',
        'portalocker',
        'prometheus_client',
        'prompt_toolkit',
        'proto-plus',
        'protobuf',
        'psutil',
        'pure_eval',
        'pyasn1',
        'pyasn1_modules',
        'pycountry',
        'pycparser',
        'pydantic',
        'pydantic_core',
        'pydash',
        'Pygments',
        'PyJWT',
        'pyodbc',
        'pyreadline3',
        'python-dateutil',
        'python-dotenv',
        'python-json-logger',
        'PyYAML',
        'pyzmq',
        'referencing',
        'regex',
        'requests',
        'requests-oauthlib',
        'rfc3339-validator',
        'rfc3986-validator',
        'rpds-py',
        'rsa',
        'safetensors',
        'scikit-learn',
        'scipy',
        'Send2Trash',
        'six',
        'sniffio',
        'soupsieve',
        'SQLAlchemy',
        'sqlalchemy-stubs',
        'stack-data',
        'strictyaml',
        'sympy',
        'tenacity',
        'termcolor',
        'terminado',
        'threadpoolctl',
        'tinycss2',
        'tokenizers',
        'tornado',
        'tqdm',
        'traitlets',
        'transformers',
        'treelib',
        'types-python-dateutil',
        'types-requests',
        'typing_extensions',
        'uri-template',
        'urllib3',
        'wcwidth',
        'webcolors',
        'webencodings',
        'websocket-client',
        'widgetsnbextension',
    ],
)